#idp_py_syntax
def flatten(l, ltypes=(list, tuple)):
    ltype = type(l)
    l = list(l)
    i = 0
    while i < len(l):
        while isinstance(l[i], ltypes):
            if not l[i]:
                l.pop(i)
                i -= 1
                break
            else:
                l[i:i + 1] = l[i]
        i += 1
    return ltype(l)

class Formula(object):
    pass

def tuple_to_atom(tup):
    try:
        typing = tup[1] + "(" + ", ".join(map(str,tup[0])) + ")"
    except TypeError:
        typing = tup[1] + "(" + tup[0] + ")"
    if len(tup) >= 3 and len(tup[2]) >= 1:
        ifs = " & ".join(map(str,tup[2]))
        return typing + " & " + ifs
    else:
        return typing

class DefRule(Formula):

    def __init__(self, args, body):
        self.args = args
        self.body = body

    def __str__(self):
        return "(" + ", ".join(self.args) + ")" + " <- " + str(self.body) + "."

class AggregateFormula(Formula):

    def __init__(self, agg, vars, formula):
        self.agg = agg
        self.vars = vars
        self.formula = formula

    def __str__(self):
        var_tuple =  flatten([x[0] for x in self.vars])
        it = (" ".join(var_tuple) + ": " +
                " & ".join(map(tuple_to_atom, self.vars)))
        if self.agg == "card":
                return self.agg + "{" + it + " }"
        return self.agg + "{" + it + " : " + str(self.formula) + "}"


class ArithmeticFormula(Formula):

    def __init__(self,symbol,children):
        self.symbol = symbol
        self.children = children

    def __str__(self):
        def add_pars(x):
            symbols =  "+-*%/^"
            return "(" + x + ")" if any([(s in x) for s in symbols]) else x
        if self.symbol == "/":
            (l,r) = [add_pars(str(x)) for x in self.children]
            return "(" + l + "-" + l + "%" + r + ") / " + r
        else:
            return (" " + self.symbol + " ").join([add_pars(str(x)) for x in self.children])

class BooleanFormula(Formula):

    def __init__(self,symbol,children):
        self.symbol = symbol
        self.children = children

    def __str__(self):
        return (" " + self.symbol + " ").join(["(" + str(x) + ")" for x in self.children])

class UnaryFormula(Formula):

    def __init__(self,symb,child):
        self.child = child
        self.symb = symb

    def __str__(self):
        return self.symb + "(" + str(self.child) + ")"

class QuantifiedFormula(Formula):

    def __init__(self, kind, vars, formula):
        self.kind = kind
        self.vars = vars
        self.formula = formula

    def guard_sym(self):
        if self.kind == "!":
            return "=>"
        if self.kind == "?":
            return "&"

    def __str__(self):
        var_tuple =  flatten([x[0] for x in self.vars])
        return (self.kind + " " + " ".join(var_tuple) + ": " +
                " & ".join(map(tuple_to_atom, self.vars)) +
                self.guard_sym() + " "+ str(self.formula))

class Comparison(Formula):

    def __init__(self,symb,left,right):
        self.symb = symb
        self.left = left
        self.right = right

    def __str__(self):
        return str(self.left) + " " + self.symb + " " + str(self.right)

import ast

class FormulaBuilder(ast.NodeVisitor):

    def __init__(self):
        pass

    def generic_visit(self, node):
        result = flatten([self.visit(child) for child in ast.iter_child_nodes(node)])
        return result

    def visit_BinOp(self, node):
        symb = self.visit(node.op)
        return ArithmeticFormula(symb, [self.visit(node.left), self.visit(node.right)])

    def visit_BoolOp(self, node):
        symb = self.visit(node.op)
        return BooleanFormula(symb, [self.visit(child) for child in node.values])

    def visit_Add(self, node): return "+"
    def visit_Sub(self, node): return "-"
    def visit_Mult(self, node): return "*"
    def visit_Div(self, node): return "/"
    def visit_Mod(self, node): return "%"
    def visit_Pow(self, node): return "^"

    def visit_Num(self, node):
        return str(node.n)

    def visit_And(self, node):
        return "&"

    def visit_Or(self, node):
        return "|"

    def visit_Not(self, node):
        return "~"

    def visit_Eq(self, node):
        return "="

    def visit_NotEq(self, node):
        return "~="

    def visit_Lt(self, node):
        return "<"

    def visit_Gt(self, node):
        return ">"

    def visit_LtE(self, node):
        return "=<"

    def visit_GtE(self, node):
        return ">="

    def visit_Compare(self, node):
        return Comparison(self.visit(node.ops[0]), self.visit(node.left), self.visit(node.comparators[0]))

    def visit_UnaryOp(self, node):
        symb = self.visit(node.op)
        return UnaryFormula(symb, self.visit(node.operand))

    def visit_Tuple(self, node):
        return tuple([self.visit(x) for x in node.elts])

    def visit_GeneratorExp(self,node):
        return self.visit_ListComp(node)

    def visit_ListComp(self,node):
        return ([(self.visit(gen.target), self.visit(gen.iter), [self.visit(an_if) for an_if in gen.ifs]) for gen in node.generators], self.visit(node.elt))

    def visit_Name(self, node):
        return node.id

    def visit_Lambda(self, node):
        return DefRule(self.visit(node.args), self.visit(node.body))

    def visit_Call(self, node):
         func = self.visit(node.func)
         if func == "any" or func == "all":
             symb = "?" if func == "any" else "!"
             return QuantifiedFormula(symb, *self.visit(node.args[0]))
         aggregates = { 'sum' : 'sum', 'len' : 'card', 'product' : 'prod', 'max' : 'max', 'min' : 'min' }
         if func in list(aggregates.keys()):
             return AggregateFormula(aggregates[func], *self.visit(node.args[0]))
         return str(func) + "(" + ", ".join(map(str, [self.visit(arg) for arg in node.args])) + ")"

def parse(string):
    import ast
    p = ast.parse(string)
    return FormulaBuilder().visit(p)[0]

def parse_formula(string):
    result = parse(string)
    return str(result)



# idp_parse_out


def parse_element(s):
    s = s.strip()
    try:
        return int(s)
    except ValueError:
        unquoted = s.strip("'").strip('"')
        return unquoted

def parse_tuple(tup):
    tup = tup.strip()
    elements = tup.split(',')
    parsed = list(map(parse_element, elements))
    if len(parsed) == 1:
        return parsed[0]
    return tuple(parsed)

def parse_function_tuple(s):
    s = s.strip()
    from_, to = s.split('->')
    return parse_tuple(from_), parse_element(to)

def parse_enumerated(s):
    if '->' in s:
        return parse_function_tuple(s)
    return parse_tuple(s)

def parse_enumeration(s):
    s = s.strip()
    elements = s.split(';')
    parsed = list(map(parse_enumerated, elements))
    if "->" in s: # Function
        return dict(parsed)
    else: # Predicate
        return parsed

def parse_range(s):
    s = s.strip()
    low, up = s.split('..')
    return list(range(int(low),int(up)))

def parse_contents(s):
    stripped = s.strip().lstrip('{').rstrip('}').strip()
    if '..' in stripped:
        return parse_range(stripped)
    return parse_enumeration(stripped)

def parse_assignment(s,wanted):
    name, contents = s.split('=')
    name = name.strip()
    if name in wanted:
        return name, parse_contents(contents)
    else:
        return None, None

def idp_parse(s,wanted):
    lines = s.splitlines()
    first_line = True
    res = {}
    for line in lines:
        line = line.strip()
        if __debug__:
            print(line)
        if first_line:
            first_line = False
            continue
        if line == "" or line.rstrip('}') == "" or "structure" in line:
            continue
        symb, content = parse_assignment(line,wanted)
        if symb:
            res[symb] = content
    return res



"""
This file contains all the IDP objects.
"""

import collections


class IDPObject(object):
    """
    'Abstract' class for all the IDP objects.
    Initialises idp.
    """
    def __init__(self, idp):
        self.idp = idp

    def accept(self, vis):
        return vis.visit(self)

# Vocabulary objects:


class IDPVocabularyObject(IDPObject):
    """
    Abstract class for all the objects that appear in a vocabulary.
    """

    def __eq__(self, other):
        """
        Two objects are equal is they have the same name.
        """
        return self.typedName() == other.typedName()

    def __init__(self, idp, name, typing):
        super(IDPVocabularyObject, self).__init__(idp)
        self.name = name
        self.typing = typing

    def in_theory(self):
        """
        Doesn't show up in a theory, so returns empty string.
        """
        return ""

    def in_vocabulary(self):
        return self.typedName()

    def in_structure(self):
        return self.name + ' = {' + self.show_contents() + '}'

    def typedName(self):
        return self.name + "(" + ", ".join(map(str, self.typing)) + ")"

    def show_tuple(self, tup):
        if isinstance(tup, tuple):
            return ', '.join(map(str, tup))
        # if the tup is already str, we need to encode it to avoid unicode
        # errors. str() would give an error here.
        else:
            return str(tup)

    def show_contents(self):
        return "; ".join([self.show_tuple(x) for x in self])

    def __len__(self):
        return len(self.to_set())

    def __contains__(self, x):
        return self.idp.normal_form(x) in self.to_set()

    def __iter__(self):
        # if (len(self.typing) > 1):
        #     return iter(self.product_of_types())
        # else:
        return iter(self.to_set())

    def product_of_types(self):
        import itertools
        types = [getattr(self.idp,x) for x in self.typing]
        return itertools.product(*types)

class IDPEnumeratedObject(IDPVocabularyObject):

    def __init__(self, idp, name, typing, enum):
        super(IDPEnumeratedObject,self).__init__(idp, name, typing)
        self._enumeration = self.idp.normal_form_enumeration(enum)

    def to_set(self):
        return set(self.enumeration)

    def changed(self):
        self.idp.dirty = True

    @property
    def enumeration(self):
        return self._enumeration
    @enumeration.setter
    def enumeration(self, enum):
        self._enumeration = enum
        self.changed()

class IDPUnknownObject(IDPEnumeratedObject):

    def in_structure(self):
        return ""

    @property
    def enumeration(self):
         self._enumeration = self.idp.fillIn(self.name)
         return self._enumeration

class IDPGeneratedObject(IDPVocabularyObject):

    def __init__(self,*args):
        super(IDPGeneratedObject,self).__init__(*args[:-1])
        self.implementation = args[-1]

    def product_of_types(self):
        import itertools
        types = [getattr(self.idp,x) for x in self.typing]
        return itertools.product(*types)

    @property
    def enumeration(self):
        return self.product_of_types()

    def to_set(self):
        return self.enumeration

# Functions


class IDPFunction(IDPVocabularyObject, collections.Mapping):

    def __init__(self, idp, name, types, return_type, partial=False):
        IDPVocabularyObject.__init__(self, idp, name, types)
        self.return_type = return_type
        self.partial = partial

    def show_tuple(self, tup):
        return super(IDPFunction,self).show_tuple(tup) + "->" + self.idp.for_idp(self[tup])

    def typedName(self):
        return ("partial " if self.partial else '') + super(IDPFunction,self).typedName() + " : " + self.return_type

    def __iter__(self):
        sup_it = super(IDPFunction, self).__iter__()
        for x in sup_it:
            if self[x] is not None:
                yield x

    def __call__(self, *args):
        return self[args]


class IDPEnumeratedFunction(IDPFunction, IDPEnumeratedObject, collections.MutableMapping):

    def __init__(self, idp, name, args, rt, enum, partial=False):
        IDPFunction.__init__(self, idp, name, args, rt, partial)
        IDPEnumeratedObject.__init__(self, idp, name, args, enum)
        # check for string

    def __setitem__(self, key, value):
        self.enumeration[self.idp.normal_form(key)] = self.idp.normal_form(value)
        self.changed()

    def __getitem__(self, key):
        result = self.enumeration[self.idp.normal_form(key)]
        if isinstance(getattr(self.idp, self.return_type), IDPIntType):
            return int(result)
        return result

    def __delitem__(self, key):
        del self.enumeration[self.idp.normal_form(key)]
        self.changed()

class IDPUnknownFunction(IDPEnumeratedFunction,IDPUnknownObject):

    def __init__(self, idp, name, args, rt, partial=False):
        super(IDPUnknownFunction, self).__init__(idp,name,args,rt,{},partial)


class IDPEmptyConstantFunction(IDPEnumeratedFunction,IDPUnknownObject):

    def __init__(self, idp, name, args, rt, partial=False):
        super(IDPEmptyConstantFunction, self).__init__(idp,name,args,rt,{},partial)

    def typedName(self):
        return self.name

class IDPValueConstantFunction(IDPEnumeratedFunction, IDPUnknownObject):

    def __init__(self, idp, name, args, rt, enumeration, partial=False):
        self._content = enumeration
        super().__init__(idp,name,args,rt,{},partial)

    def in_structure(self):
        return self.name + " = " + str(self._content)



class IDPGeneratedFunction(IDPFunction, IDPGeneratedObject):

    def __init__(self, idp, name, args, rt, impl, partial=False):
        IDPFunction.__init__(self,idp,name,args,rt, partial)
        IDPGeneratedObject.__init__(self,idp,name,args,impl)

    def __getitem__(self, key):
        args = list(map(self.idp.object_for_name, key))
        try:
            res = self.implementation(args)
        except AttributeError:
            if len(key) == 1:
                res = self.implementation(args[0])
        return self.idp.normal_form(res)


# class IDPGeneratedPartialFunction(IDPPartialFunction, IDPGeneratedFunction):

#     def __getitem__(self, key):
#         res = super(IDPGeneratedPartialFunction, self).__getitem__(key)
#         if self.idp.assign_name(None) == res:
#             return None
#         else:
#             return res

#### Predicates

class IDPPredicate(IDPVocabularyObject, collections.Set):

    def __call__(self, *args):
        return self.idp.normal_form(args) in self

class IDPUnknownPredicate(IDPPredicate, IDPUnknownObject):

    def __init__(self, idp, name, types):
        IDPPredicate.__init__(self,idp,name,types)

class IDPEnumeratedPredicate(IDPPredicate, IDPEnumeratedObject, collections.MutableSet):

    def add(self, x):
        self.enumeration.append(self.idp.normal_form(x))
        self.changed()

    def discard(self, x):
        self.enumeration.discard(x)
        self.changed()

class IDPGeneratedPredicate(IDPPredicate, IDPGeneratedObject):

    pass

class IDPType(IDPEnumeratedPredicate):

    def __init__(self, idp, name, enum):
        super(IDPType, self).__init__(idp, name, [], enum)

    def typedName(self):
        return self.name

    def in_vocabulary(self):
        return "type " + self.name

   # def isIntType(self):
    #    x = self.enum[0]
     #   return int(x) == x

class IDPIntType(IDPType):

    def in_vocabulary(self):
        return super(IDPIntType, self).in_vocabulary() + " isa int"


class IDPIntRangeType(IDPIntType):

    def in_vocabulary(self):
        return super(IDPIntRangeType, self).in_vocabulary()

    def show_contents(self):
        return "".join(self._enumeration)


class IDPConstructedType(IDPType):

    def in_vocabulary(self):
        return super(IDPConstructedType, self).in_vocabulary() +\
            " constructed from {" + self.show_constructed_from() + "}"

    def show_constructed_from(self):
        content = ", ".join(self._enumeration)
        return content

    # No need for anything in struct
    def in_structure(self):
        return ""


class IDPSpecialType(IDPType):

    def in_vocabulary(self):
        return super().in_vocabulary() + " isa " + "".join(self._enumeration)

    # No need for anything in struct
    def in_structure(self):
        return ""

##### Theory objects

class IDPTheoryObject(IDPObject):

    def in_structure(self):
        return ""

    def in_vocabulary(self):
        return ""

    def in_theory(self):
        raise NotImplementedError("All theory objects should implement the 'in_theory' method")

class IDPConstraint(IDPTheoryObject):

    def __init__(self, idp, formula):
        super(IDPConstraint,self).__init__(idp)
        self.formula = formula

    def in_theory(self):
        if self.formula[-1] != ".":
            return self.formula + "."
        else:
            return self.formula

class IDPRule(IDPTheoryObject):

    def __init__(self, idp, head_pred, vars_, body):
        super(IDPRule,self).__init__(idp)
        self.head_pred = head_pred
        self.vars = vars_
        self.body = body

    def in_theory(self):
        return  "!" + " ".join(self.vars) + ": " + self.head_pred + "(" + (", ").join(self.vars) + ") <- " + self.body + "."

class IDPRuleStr(IDPTheoryObject):

    def __init__(self, idp, string):
        super(IDPRuleStr,self).__init__(idp)
        self.rule = string

    def in_theory(self):
        return self.rule

class IDPDefinition(IDPTheoryObject):

    def __init__(self, idp, rule_list):
        super(IDPDefinition, self).__init__(idp)
        self.rules = []
        for rule in rule_list:
            if isinstance(rule, str):
                self.rules.append(rule)
            else:
                self.rules.append(rule.in_theory())

    def in_theory(self):
        theory_str = "{\n" + "\n".join([x for x in self.rules])
        if theory_str[-1] != ".":
            theory_str += "."
        return theory_str + "\n}"
